"""Identification of outliers based on autoencoder reconstruction error.

<p>This check identifies periods of anomalies based autoencoded deep neural
network architecture. It is based on the assumption that normal data is easy
to reconstruct (encode) while anomalous data will have a higher error.</p>
<p><img src='../static/images/reporting/multivariate_anomaly.svg'></p>
<p class="scoring-explanation">The score of this check is calculated based on the count of all
points where an autoencoder anomaly occurs. Imagine that 100 points are analyzed in a given time-frame
and there are 10 autoencoder anomalies. The score for this check in that case would be
90% = 1 - 10 / 100. Which means that for 90% of all points no autoencoder anomaly occurs.</p>
<div class="ts-check-impact">
<p>The presence of anomalies detected by the autoencoder might indicate process upsets as well
as istrumentation issues.</p>
</div>
"""

from datetime import timedelta

import numpy as np
import pandas as pd

from sklearn.neural_network import MLPRegressor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import Normalizer, MinMaxScaler


import timeseer

from timeseer import DataType
from timeseer.analysis.utils import event_frames_from_dataframe


_CHECK_NAME = "Auto-encoder outlier"
_EVENT_FRAME_NAME = "Auto-encoder outlier"

META = {
    "checks": [
        {
            "name": _CHECK_NAME,
            "kpi": "Validity",
            "type": "smell",
            "event_frames": [_EVENT_FRAME_NAME],
        },
    ],
    "conditions": [
        {
            "min_series": 3,
            "min_weeks": 1,
            "min_data_points": 300,
            "data_type": [
                DataType.FLOAT32,
                DataType.FLOAT64,
                DataType.DICTIONARY,
                DataType.CATEGORICAL,
            ],
        }
    ],
    "signature": "multivariate",
}


def _calculate_total_check_active_time(frames, median_archival_step):
    active_time = timedelta(0)
    for _, row in frames.iterrows():
        length = row["end_date"] - row["start_date"]
        if length == timedelta(0):
            length = timedelta(seconds=median_archival_step)
        active_time = active_time + length
    return active_time


def _detect_mad_outliers(points):
    median = np.median(points, axis=0)
    deviation = np.abs(points - median)
    med_abs_deviation = np.median(deviation)
    modified_z_score = 0.6745 * deviation / med_abs_deviation

    return modified_z_score


def _get_intervals(df, anomalies_mse, event_type):
    anomalies = pd.Series(data=anomalies_mse, index=df.index)
    interval_grp = (anomalies != anomalies.shift().bfill()).cumsum()

    intervals = (
        df.assign(interval_grp=interval_grp)[anomalies]
        .reset_index()
        .groupby(["interval_grp"])
        .agg(start_date=("ts", "first"), end_date=("ts", "last"))
    )
    intervals["type"] = event_type
    return intervals


def _get_active_points(concatenated_df):
    pipeline = Pipeline([("normalizer", Normalizer()), ("scaler", MinMaxScaler())])
    pipeline.fit(concatenated_df)
    train_data = pipeline.transform(concatenated_df)
    n_latent = int(np.ceil(np.log2(concatenated_df.shape[1])))
    reg = MLPRegressor(
        hidden_layer_sizes=(n_latent),
        activation="tanh",
        solver="adam",
        learning_rate_init=0.0001,
        max_iter=20,
        tol=0.0000001,
        verbose=False,
    )
    reg.fit(train_data, train_data)

    reconstructions = reg.predict(train_data)
    mse = np.mean(np.power(train_data - reconstructions, 2), axis=1)

    threshold = 3.5
    mad_z_scores = _detect_mad_outliers(mse)
    return mad_z_scores > threshold


def _get_mv_median_archival_step(df):
    meas_times = df.index.to_series()
    diff_times = meas_times - meas_times.shift()
    diff_times.dropna()
    return pd.Timedelta(np.median(diff_times)).total_seconds()


def _clean_input(inputs):
    return (
        pd.concat(
            [
                series.data[~series.data.index.duplicated(keep="first")]
                for series in inputs
            ],
            axis=1,
            sort=False,
        )
        .interpolate("linear")
        .dropna()
        .sort_index()
    )


def _run_autoencoder_anomaly_detection(inputs):
    concatenated_df = _clean_input(inputs)
    median_archival_step = _get_mv_median_archival_step(concatenated_df)

    active_points = _get_active_points(concatenated_df)
    intervals = _get_intervals(concatenated_df, active_points, _EVENT_FRAME_NAME)
    check_active_time = _calculate_total_check_active_time(
        intervals, median_archival_step
    )
    frames = event_frames_from_dataframe(intervals)

    total_time = concatenated_df.index[-1] - concatenated_df.index[0]
    percentage = check_active_time / total_time

    return percentage, list(frames)


def run(
    analysis_input: timeseer.MultivariateAnalysisInput,
):  # pylint: disable=missing-function-docstring
    inputs = analysis_input.inputs

    percentage, event_frames = _run_autoencoder_anomaly_detection(inputs)

    check = timeseer.CheckResult(_CHECK_NAME, float(percentage))

    return timeseer.AnalysisResult(check_results=[check], event_frames=event_frames)
